package no.vegvesen.meldingsformidler.common.openamclient

import com.nimbusds.oauth2.sdk.AuthorizationGrant
import com.nimbusds.oauth2.sdk.ParseException
import com.nimbusds.oauth2.sdk.ResourceOwnerPasswordCredentialsGrant
import com.nimbusds.oauth2.sdk.Scope
import com.nimbusds.oauth2.sdk.TokenRequest
import com.nimbusds.oauth2.sdk.auth.ClientAuthentication
import com.nimbusds.oauth2.sdk.auth.ClientSecretBasic
import com.nimbusds.oauth2.sdk.auth.Secret
import com.nimbusds.oauth2.sdk.http.HTTPResponse
import com.nimbusds.oauth2.sdk.id.ClientID
import com.nimbusds.openid.connect.sdk.OIDCTokenResponse
import com.nimbusds.openid.connect.sdk.OIDCTokenResponseParser
import java.io.IOException
import java.net.URI
import no.vegvesen.meldingsformidler.common.openamclient.OpenAmClient.BearerToken

class Oauth2Client internal constructor(
    private val password: Secret,
    private val clientID: ClientID,
    private val clientSecret: Secret,
    private val clientAuth: ClientAuthentication,
    private val passwordGrant: AuthorizationGrant,
    private val scope: Scope,
    private val tokenRequest: TokenRequest,
) {

    companion object {
        fun create(
            openAmUrl: String,
            openAmUsername: String,
            openAmPassord: String,
            oidcClientId: String,
            oidcClientSecret: String,
        ): Oauth2Client {
            val password = Secret(openAmPassord)
            val clientID = ClientID(oidcClientId)
            val clientSecret = Secret(oidcClientSecret)
            val clientAuth: ClientAuthentication = ClientSecretBasic(clientID, clientSecret)
            val passwordGrant: AuthorizationGrant = ResourceOwnerPasswordCredentialsGrant(openAmUsername, password)
            val scope = Scope("openid", "svvprofile")
            val tokenRequest = TokenRequest(URI.create(openAmUrl), clientAuth, passwordGrant, scope)

            return Oauth2Client(
                password,
                clientID,
                clientSecret,
                clientAuth,
                passwordGrant,
                scope,
                tokenRequest,
            )
        }
    }

    @Throws(IOException::class, ParseException::class, RuntimeException::class)
    fun getOrThrow(): BearerToken {
        val send: HTTPResponse = tokenRequest.toHTTPRequest().send()
        val response = OIDCTokenResponseParser.parse(send)

        if (!response.indicatesSuccess()) {
            val errorResponse = response.toErrorResponse()
            throw RuntimeException("Failed to fetch OIDC access token. Error message: " + errorResponse.toErrorResponse().errorObject.toString())
        }

        val successResponse = response.toSuccessResponse() as OIDCTokenResponse
        val accessToken = successResponse.oidcTokens.accessToken

        return BearerToken(accessToken.value)
    }
}
