package no.vegvesen.meldingsformidler.common.openamclient

import kotlin.time.Duration
import kotlin.time.Duration.Companion.minutes
import kotlin.time.Duration.Companion.seconds
import kotlin.time.toKotlinDuration
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.EmptyToken
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.Expired
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.ShouldFetchNewToken
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.Valid
import org.slf4j.LoggerFactory

/**
 * @param timeLeftToBeConsideredAlmostExpired When the token expiration time is below this treshold, it will try to
 * fetch a new token, or use the existing one if it fails.
 * @param timeLeftToBeConsideredExpired When the token expiration time is less than this threshold,
 * the token is considered expired.
 */
class TokenValidator(
    private val timeLeftToBeConsideredAlmostExpired: Duration = 15.minutes,
    private val timeLeftToBeConsideredExpired: Duration = 30.seconds,
) {

    private val log = LoggerFactory.getLogger(this::class.java)

    companion object {
        @JvmStatic
        fun create(): TokenValidator {
            return TokenValidator()
        }

        @JvmStatic
        fun create(
            timeLeftToBeConsideredAlmostExpired: java.time.Duration,
            timeLeftToBeConsideredExpired: java.time.Duration,
        ): TokenValidator = TokenValidator(
            timeLeftToBeConsideredAlmostExpired.toKotlinDuration(),
            timeLeftToBeConsideredExpired.toKotlinDuration(),
        )
    }

    enum class Result {
        EmptyToken,
        Valid,
        ShouldFetchNewToken,
        Expired
    }


    fun validate(bearerToken: OpenAmClient.BearerToken): Result {
        if (bearerToken.token.isEmpty()) {
            log.debug("Token validator: No token was found within bearer token.")
            return EmptyToken
        }
        val isLongTimeUntilTokenExpires = bearerToken.remainingValidTime() > timeLeftToBeConsideredAlmostExpired
        if (isLongTimeUntilTokenExpires) {
            log.debug("Token validator: Token is valid, it expires in ${bearerToken.remainingValidTime().inWholeMinutes} minutes. Token id: ${bearerToken.id()}.")

            return Valid
        }

        val tokenExpiresSoon = bearerToken.remainingValidTime() > timeLeftToBeConsideredExpired
        if (tokenExpiresSoon) {
            log.debug("Token validator: Token is soon invalid, a new should be fetched. It expires in ${bearerToken.remainingValidTime().inWholeMinutes} minutes. Token id: ${bearerToken.id()}.")
            return ShouldFetchNewToken
        }

        log.debug("Token validator: Token is expired. Token id: ${bearerToken.id()}.")
        return Expired
    }
}
