@file:Suppress("MemberVisibilityCanBePrivate")

package no.vegvesen.meldingsformidler.common.securetext

import java.security.KeyFactory
import java.security.PrivateKey
import java.security.PublicKey
import java.security.spec.PKCS8EncodedKeySpec
import java.security.spec.X509EncodedKeySpec
import java.util.Base64
import javax.crypto.Cipher

object SecureText {

    /**
     * Converts a Base64 encoded public key string to a PublicKey object.
     * The input string must be in the PEM format with header "-----BEGIN PUBLIC KEY-----"
     * and footer -----END PUBLIC KEY-----".
     *
     * @param base64PublicKey The Base64 encoded public key string in PEM format.
     * @return The PublicKey object.
     * @throws IllegalArgumentException If the provided string is not a valid Base64 encoded public key.
     */
    fun getPublicKey(base64PublicKey: String): PublicKey {
        val cleanKey = base64PublicKey
            .replace("-----BEGIN PUBLIC KEY-----", "")
            .replace("-----END PUBLIC KEY-----", "")
            .replace("\n", "")
        val keyBytes = Base64.getDecoder().decode(cleanKey)

        val spec = X509EncodedKeySpec(keyBytes)
        val keyFactory = KeyFactory.getInstance("RSA")

        return keyFactory.generatePublic(spec)
    }

    /**
     * Converts a Base64 encoded private key string to a PrivateKey object.
     * The input string must be in the PEM format with header "-----BEGIN RSA PRIVATE KEY-----"
     * and footer "-----END RSA PRIVATE KEY-----".
     *
     * @param base64PrivateKey The Base64 encoded private key string in PEM format.
     * @return The PrivateKey object.
     */
    fun getPrivateKey(base64PrivateKey: String): PrivateKey {
        val cleanKey = base64PrivateKey
            .replace("-----BEGIN RSA PRIVATE KEY-----", "")
            .replace("-----END RSA PRIVATE KEY-----", "")
            .replace("\n", "")
        val keyBytes = Base64.getDecoder().decode(cleanKey)

        val keySpec = PKCS8EncodedKeySpec(keyBytes)
        val keyFactory = KeyFactory.getInstance("RSA")

        return keyFactory.generatePrivate(keySpec)
    }

    /**
     * Encrypts a plain text string using the provided public key and returns the result as a single string.
     * The text is split into chunks of the specified size before encryption. If an exception is thrown because
     * the string is too long, choose a smaller chunk size.
     *
     * @param publicKey The public key used for encryption.
     * @param plainText The plain text string to be encrypted.
     * @param chunkSize The size of each chunk to be encrypted. Defaults to 300.
     * @return The encrypted text chunks as a single string.
     * @throws IllegalArgumentException If the encryption process fails.
     */
    fun encrypt(publicKey: PublicKey, plainText: String, chunkSize: Int = 300): String =
        encryptToList(publicKey, plainText, chunkSize)
            .toString()

    /**
     * Encrypts a plain text string using the provided public key and returns the result as a list of encrypted chunks.
     * The text is split into chunks of the specified size before encryption. If an exception is thrown because
     * the string is too long, choose a smaller chunk size.
     *
     * @param publicKey The public key used for encryption.
     * @param plainText The plain text string to be encrypted.
     * @param chunkSize The size of each chunk to be encrypted. Defaults to 300.
     * @return A list of encrypted text chunks.
     * @throws IllegalArgumentException If the encryption process fails.
     */
    fun encryptToList(publicKey: PublicKey, plainText: String, chunkSize: Int = 300): List<String> =
        splitTextInChunks(plainText, chunkSize)
            .map { encryptString(publicKey, it) }

    /**
     * Decrypts a comma-separated string of encrypted text chunks using the provided private key.
     * The input string must be in the format of an array starting with '[',
     * with every element split by ', ', and ending with ']'.
     *
     * @param privateKey The private key used for decryption.
     * @param encryptedTextArray The comma-separated string of encrypted text chunks in the array format.
     * @return The decrypted string.
     * @throws IllegalArgumentException if the input string does not contain '[' and ']'.
     */
    fun decryptToString(privateKey: PrivateKey, encryptedTextArray: String): String {
        val hasArrayCharacters = encryptedTextArray.contains("[") && encryptedTextArray.contains("]")
        require(hasArrayCharacters) {
            "Input string must be an array of encrypted data chunks in an array starting with '[' , every element split by ', ' and ending with ']'"
        }

        return encryptedTextArray
            .trimIndent()
            .replace("\n", "")
            .replace("[", "")
            .replace("]", "")
            .split(", ")
            .joinToString("") { decryptString(privateKey, it) }
    }

    /**
     * Decrypts a list of encrypted text chunks using the provided private key and returns the result as a list of decrypted strings.
     *
     * @param privateKey The private key used for decryption.
     * @param encryptedText A list of encrypted text chunks.
     * @return A list of decrypted strings.
     * @throws IllegalArgumentException If the decryption process fails.
     */
    fun decryptToList(privateKey: PrivateKey, encryptedText: List<String>): List<String> =
        encryptedText
            .map { decryptString(privateKey, it) }


    private fun splitTextInChunks(input: String, chunkSize: Int): MutableList<String> {
        val length = input.length
        val chunks = mutableListOf<String>()
        for (i in 0 until length step chunkSize) {
            val end = (i + chunkSize).coerceAtMost(length)
            val chunk = input.substring(i, end)
            chunks.add(chunk)
        }
        return chunks
    }

    private fun encryptString(publicKey: PublicKey, plainText: String): String {

        val cipher = Cipher.getInstance("RSA")
        cipher.init(Cipher.ENCRYPT_MODE, publicKey)

        val encryptedBytes = cipher.doFinal(plainText.toByteArray(Charsets.UTF_8))

        return Base64.getEncoder().encodeToString(encryptedBytes)
    }

    fun decryptString(privateKey: PrivateKey, encryptedText: String): String {
        val cipher = Cipher.getInstance("RSA")
        cipher.init(Cipher.DECRYPT_MODE, privateKey)

        val decryptedBytes = cipher.doFinal(Base64.getDecoder().decode(encryptedText))

        return String(decryptedBytes, Charsets.UTF_8)
    }

}
