package no.vegvesen.meldingsformidler.common.securetext

import java.security.PrivateKey
import org.bouncycastle.jce.provider.BouncyCastleProvider
import org.junit.jupiter.api.Test
import java.security.PublicKey
import java.security.Security
import org.assertj.core.api.Assertions.assertThat
import org.junit.jupiter.api.BeforeAll
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.assertThrows

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class SecureTextTest {

    private lateinit var publicKeyRsa: PublicKey
    private lateinit var privateKeyRsa: PrivateKey

    @BeforeAll
    fun beforeAll() {
        Security.addProvider(BouncyCastleProvider())

        publicKeyRsa = SecureText.getPublicKey(
            """
            -----BEGIN PUBLIC KEY-----
            MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCNg8U+OkXe687QO2OgEaucV26k
            WqRoARVfF+UA5AuQSInMHC9xOMEHCFeMVrt2NEBHsFZiTl46bJH5voHh0z35cKfO
            OS+BBPiuniC8r0npvoL1e1oiO9AqqPDAuyyGyK22tqWLiCqEWZA/XG7fZhHpv7+N
            lpWBFazwKlqrvDHlWQIDAQAB
            -----END PUBLIC KEY-----
        """.trimIndent()
        )

        privateKeyRsa = SecureText.getPrivateKey(
            """
            -----BEGIN RSA PRIVATE KEY-----
            MIICXQIBAAKBgQCNg8U+OkXe687QO2OgEaucV26kWqRoARVfF+UA5AuQSInMHC9x
            OMEHCFeMVrt2NEBHsFZiTl46bJH5voHh0z35cKfOOS+BBPiuniC8r0npvoL1e1oi
            O9AqqPDAuyyGyK22tqWLiCqEWZA/XG7fZhHpv7+NlpWBFazwKlqrvDHlWQIDAQAB
            AoGAIB2McM/WtX18ZONle5knAgcRk/kCIJxGVltc3DPkr+zWYNzwo1hG7iqle7Lf
            Pjcjc2GAaTTxy36CLlRes3o2MQQh7ebQM0wNpvw/gjh4apmKdYGOH+3lRP8xJBY7
            WkxZ0fqNxxmJX+nKxymjLL+dKLXr+HM41JgPbrH6uolH/xUCQQDKkhF45YqKJ1zx
            +DqL4+AXTDJl0fFK+zqaGGYKUvSeATU3hwdgAKphqa1YTp/OVwHkvW1l93cmwp7x
            3ohke/tvAkEAstcZzkflOoNUOL2YFZ4ht4NW1cib3SMrgG1oqS7MMd3ZHlIkZh61
            D+ZqjuLyy2unetBOgVJKStQ8644vW1DntwJABFrpl0DQHrJD9EvS7mkCy11vA/Nk
            jakjFGcL4pLwAWmjPENh7DV3dEwSlwBxvRzvxHwyGuwcGrA6fJa9NfPYbwJBAKTZ
            deYZi4kGKYor6ZZzruyz2kUxp8qNcrntH/5jLsIAlvRGD1s6jvfbEvom5FryHu0s
            OlEmg2i6fnU7GRf5Wb0CQQDFgt6EELRravdDuT99iOR5Slo2Hkv3RTFJ1r0/DE4A
            WSEKh5KEhruBO25wzW74+l6GBYxzkZPECrJhFRaoib7Q
            -----END RSA PRIVATE KEY-----
        """.trimIndent()
        )

    }

    @Test
    fun `successfully encrypt a plain text string`() {
        val encryptedChunks: String = SecureText.encrypt(
            publicKey = publicKeyRsa,
            plainText = "The text to encrypt - add it to this text block <3",
            chunkSize = 30
        )

        println(encryptedChunks)
    }

    @Test
    fun `successfully decrypt an array of encrypted chunks`() {
        val expected = "The text to encrypt - add it to this text block <3"
        val chunks =
            "[YBtRexPla+uwnG9tuDbeid8u2gWa3pK+WLsXjlZJDsDjs+CBwXKvJSe359gd6QBZFYkWsmsJeeU9ULPU6APsgSOxX5+JOUuD9BNu+ZW5pu0FTNk+yfCckSmO79heu2prwnnWvuECzRaHqXoBhXrrjMZyQLMWDxTuQ5J2obmOaYg=, " +
                    "cMBKzrFDbl3pE8vsP9yP+lZO+u8tMleamZMrSO3Mpqh9UzBiudnKE5mxixsHmiDU9jUKC2GydokuhdxAQoDbbUzKcSRB9kt4hIS51TZuakaSxKRd0Vgho5vPkbbXwXqN1NdyPTAcEBErkOYi8joLohbZgov/zG/aHQ36poJRF6k=]"

        val actual = SecureText.decryptToString(privateKeyRsa, chunks)

        assertThat(actual).contains(expected)

        println("Decrypted text: \n\"$actual\"")
    }

    @Test
    fun `decrypt invalid array string throws illegal argument exception`() {
        val invalidChunkBecauseNotAnArrayInAString =
            """
            
                bidjYr3CY88BhI5/y1SufPOC77qiychGSQ2DWeiLiemkdGS4emHtqzwy1mbsPIvws+Vlg+b1M1QKhkix7i8dJ/fc2TB404yQ8zDBZxIDb9ocL7uc994p2XR02/8+SUFjMKgwB4iBvJYVte/XzliqJmOpk2sPMSZKIZywo92aHMc=

            """

        val message = assertThrows<IllegalArgumentException> {
            SecureText.decryptToString(privateKeyRsa, invalidChunkBecauseNotAnArrayInAString)
        }.message

        assertThat(message).contains("Input string must be an array of encrypted data chunks in an array starting with '[' , every element split by ', ' and ending with ']'")
    }

    @Test
    fun `encrypt and decrypt text and verify that it is unchanged`() {
        val textToEncrypt = "The secret text that nobody should see."
        val expected = listOf(textToEncrypt)

        val encryptedList: List<String> = SecureText.encryptToList(publicKeyRsa, textToEncrypt)

        val actual = SecureText.decryptToList(privateKeyRsa, encryptedList)

        assertThat(actual).isEqualTo(expected)
    }
}
